"""Client facilitating calendar item management in Exchange.

This client is a wrapper around exchangelib, and offers:

- Creating calendar items.
- Updating calendar items.
- Deleting calendar items.

When an item is created, an additional identifier is stored in the calendar
item trough the use of ExtendedProperties.

The identifier is permanent, in contrast
to the id-property set on calendar items by exchange (rumour has it these can
change without notice, due to a multitude of reasons).

Subsequent updates and deletions of calendar items are performed by searching
the account used for storage, and performing the operation on ALL calendar
items found.

All calendar items with a matching identity is operated upon, since users are
free to alter calendar items in their client of choice.
"""
from typing import List, Optional
import datetime
import uuid

from exchangelib import (
    CalendarItem,
    Account,
    FaultTolerance,
    Credentials,
    Configuration,
    BASIC,
    IMPERSONATION,
    DELEGATE,
    ExtendedProperty,
    EWSDateTime,
    UTC,
    HTMLBody,
    errors as exchange_errors,
)
from exchangelib.items import SEND_TO_NONE

from . import errors


class EwsClient(object):
    def __init__(
        self,
        identity_property_name: str,
        server: str,
        username: str,
        password: str,
        impersonation: bool = True,
        service_mode: bool = True,
        assume_access_to_calendar_only: bool = True,
    ) -> None:

        self.impersonation_access = impersonation
        self.assume_access_to_calendar_only = assume_access_to_calendar_only

        self.credentials = Credentials(username=username, password=password)
        self.identity_property_name = identity_property_name

        params = {}
        if service_mode:
            params["retry_policy"] = FaultTolerance()

        self.config = Configuration(
            server=server, credentials=self.credentials, auth_type=BASIC, **params
        )

        ns = uuid.UUID("7ee114a2-0646-5bdf-b578-f0bceb362b2d")
        identity = uuid.uuid5(ns, identity_property_name)

        class Identificator(ExtendedProperty):
            property_set_id = str(identity)
            property_name = identity_property_name
            property_type = "String"

        # Register Identificator to a variable on the calendar item named as
        # the value of `identity_property_name`, so we can set a unique
        # identificator on each CalendarItem, for multiple integrations
        try:
            CalendarItem.register(identity_property_name, Identificator)
        except ValueError:
            # If we wind up here, the CalendarItem class has most likely been
            # kept in memory.
            pass

    def get_user_calendar(self, address: str) -> Account:
        """
        Fetch the parent object needed to operate on a users calendar.

        For exchangelib this is a configured exchangelib.Account object.
        """
        account = Account(
            primary_smtp_address=address,
            credentials=self.credentials,
            config=self.config,
            access_type=(IMPERSONATION if self.impersonation_access else DELEGATE),
        )
        if not self.assume_access_to_calendar_only:
            return account
        else:
            from exchangelib.folders import Root, Calendar

            root = Root(account=account)
            cal = Calendar(is_distinguished=True, name="calendar", root=root)
            account.root = root
            account.calendar = cal
            return account

    def __convert_datetime(self, dt: datetime.datetime) -> EWSDateTime:
        r = EWSDateTime.from_datetime(
            dt.astimezone(datetime.timezone.utc).replace(tzinfo=UTC)
        )
        return r

    def __prepare_html(self, body):
        return HTMLBody("<html><body>{body}</body></html>".format(body=body))

    def create_item(
        self,
        user_calendar: Account,
        identificator: str,
        subject: str,
        body: str,
        start: datetime.datetime,
        end: datetime.datetime,
        location: Optional[str] = None,
        cancelled: bool = False,
    ) -> CalendarItem:
        if not start.tzinfo or not end.tzinfo:
            raise errors.AwareDateTimeRequired()

        item = {
            "folder": user_calendar.calendar,
            "subject": subject,
            "body": self.__prepare_html(body),
            "start": self.__convert_datetime(start),
            "end": self.__convert_datetime(end),
            self.identity_property_name: identificator,
        }
        if location:
            item["location"] = location

        if cancelled:
            item["is_cancelled"] = True
            item["legacy_free_busy_status"] = "Free"

        event = CalendarItem(**item)
        try:
            event.save(send_meeting_invitations=SEND_TO_NONE)
        except exchange_errors.ErrorFolderNotFound:
            raise errors.AccessDenied("save")
        return event

    def update_item(
        self,
        user_calendar: Account,
        identificator: str,
        subject: str,
        body: str,
        start: datetime.datetime,
        end: datetime.datetime,
        location: Optional[str] = None,
        cancelled: bool = False,
    ) -> List[CalendarItem]:
        if not start.tzinfo or not end.tzinfo:
            raise errors.AwareDateTimeRequired()

        r = list()
        try:
            for x in user_calendar.calendar.filter(
                **{self.identity_property_name: identificator}
            ):
                x.subject = subject
                x.location = location
                x.body = self.__prepare_html(body)
                x.start = self.__convert_datetime(start)
                x.end = self.__convert_datetime(end)
                x.is_cancelled = True if cancelled else False
                x.legacy_free_busy_status = "Free" if cancelled else "Busy"

                try:
                    x.save()
                except exchange_errors.ErrorAccessDenied:
                    raise errors.AccessDenied("save")
                r.append(x)
        # TODO: Remove 'except AttributeError…' in next release (> v2.0.1 ) of exchangelib
        #       Should probably check if elements returned from filter is instance of Exception
        except AttributeError as e:
            if str(e) == "'ErrorItemNotFound' object has no attribute 'get'":
                raise errors.AccessDenied("search")
            else:
                raise e
        except exchange_errors.ErrorItemNotFound:
            raise errors.AccessDenied("search")
        return r

    def delete_item(
        self, user_calendar: Account, identificator: str
    ) -> List[CalendarItem]:
        r = []
        try:
            for x in user_calendar.calendar.filter(
                **{self.identity_property_name: identificator}
            ):
                try:
                    x.delete()
                except exchange_errors.ErrorAccessDenied:
                    raise errors.AccessDenied("delete")
                r.append(x)
        except AttributeError as e:
            if str(e) == "'ErrorItemNotFound' object has no attribute 'get'":
                raise errors.AccessDenied("search")
            else:
                raise e
        # TODO: Remove 'except AttributeError…' in next release (> v2.0.1 ) of exchangelib
        #       Should probably check if elements returned from filter is instance of Exception
        except exchange_errors.ErrorItemNotFound:
            raise errors.AccessDenied("search")
        return r

    def get_all_items(self, user_calendar: Account) -> List[CalendarItem]:
        return list(
            user_calendar.calendar.filter(
                **{f"{self.identity_property_name}__exists": True}
            )
        )

    def get_identificator_from_event(self, event: CalendarItem) -> str:
        person_event_id = getattr(event, self.identity_property_name)
        assert isinstance(person_event_id, str)
        assert "-" in person_event_id
        return person_event_id
